
; The usual stream operations

(define-syntax cons-stream
  (syntax-rules ()
    ((cons-stream head tail)
     (cons head (delay tail)))))

(define head car)
(define (tail s) (force (cdr s)))
(define the-empty-stream '<the-empty-stream>)
(define (empty-stream? s)
  (eq? s the-empty-stream))

(define (stream s)
  (cons-stream s the-empty-stream))


; Map a procedure across a stream

(define (map-stream proc . streams)
  (let loop ((streams streams))
    (if (some empty-stream? streams)
	the-empty-stream
	(cons-stream (apply proc (map head streams))
		     (loop (map tail streams))))))

(define (some pred l)
  (and (not (null? l))
       (or (pred (car l)) (some pred (cdr l)))))

(define (filter-stream proc stream)
  (if (empty-stream? stream)
      the-empty-stream
      (if (proc (head stream))
	  (cons-stream (head stream) (filter-stream proc (tail stream)))
	  (filter-stream proc (tail stream)))))

(define (append-streams stream1 stream2-thunk)
  (if (empty-stream? stream1)
      (stream2-thunk)
      (cons-stream (head stream1)
		   (append-streams (tail stream1) stream2-thunk))))

(define (stream-prefix stream n)
  (if (empty-stream? stream)
      the-empty-stream
      (if (= n 0)
	  the-empty-stream
	  (cons-stream (head stream)
		       (stream-prefix (tail stream) (- n 1))))))

(define (stream-tail stream n)
  (if (empty-stream? stream)
      the-empty-stream
      (if (= n 0)
	  stream
	  (stream-tail (tail stream) (- n 1)))))

(define (list->stream l)
  (if (null? l)
      the-empty-stream
      (cons-stream (car l) (list->stream (cdr l)))))


; Finite merge.

(define-syntax merged-stream
  (syntax-rules ()
    ((merged-stream s1 ...)
     (merge-streams (list (lambda () s1) ...)))))

(define (merge-streams thunks)
  (merge-streams-loop thunks '()))

(define (merge-streams-loop thunks done)
  (if (null? thunks)
      (if (null? done)
	  the-empty-stream
	  (merge-streams (reverse done)))
      (let ((s1 ((car thunks))))
	(if (empty-stream? s1)
	    (merge-streams-loop (cdr thunks) done)
	    (cons-stream (head s1)
			 (merge-streams-loop (cdr thunks)
					     (cons (lambda () (tail s1))
						   done)))))))

; Infinite merge.

(define (merge-stream-of-streams streams)
  (diagonalize streams 1))

(define (diagonalize streams n)
  (let ((streams (filter-stream (lambda (s) (not (empty-stream? s)))
				streams)))
    (if (empty-stream? streams)
	the-empty-stream
	(let ((prefix (stream-prefix streams n)))
	  (append-streams
	       (map-stream head prefix)
	       (lambda ()
		 (diagonalize (append-streams (map-stream tail prefix)
					      (lambda ()
						(stream-tail streams n)))
			      (+ n 1))))))))

; Form the stream of all applications of proc to the members of s1 and s2.

(define-syntax product-stream
  (syntax-rules ()
    ((product-stream proc stream ...)
     (%product-stream proc (list (lambda () stream) ...)))))

(define (%product-stream proc thunks)
  (let loop ((thunks thunks) (streams '()))
    (if (null? thunks)
	(let recur ((streams streams) (args '()))
	  (if (null? streams)
	      (stream (apply proc args))
	      (merge-stream-of-streams
	       (map-stream (lambda (arg)
			     (recur (cdr streams) (cons arg args)))
			   (car streams)))))
	(let ((stream ((car thunks))))
	  (if (empty-stream? stream)
	      the-empty-stream
	      (loop (cdr thunks) (cons stream streams)))))))


; Remove duplicates.

(define (without-duplicates stream)
  (without-duplicates-loop stream '()))

(define (without-duplicates-loop stream seen)
  (if (empty-stream? stream)
      stream
      (let ((x (head stream)))
	(if (member x seen)
	    (without-duplicates-loop (tail stream) seen)
	    (cons-stream x (without-duplicates-loop (tail stream)
						    (cons x seen)))))))


; Output utility: display a stream as {x y z ...}

(define (show-stream s)
  (write-stream s)
  (newline))

(define (write-stream s)
  (let recur ((s s) (n *show-stream*))
    (cond ((stream? s)
	   (display "{")
	   (if (not (empty-stream? s))
	       (begin (recur (head s) (quotient n 2))
		      (let loop ((n n) (s (tail s)))
			(if (not (empty-stream? s))
			    (if (<= n 0)
				(display "...")
				(begin (display " ")
				       (recur (head s) (quotient n 2))
				       (loop (- n 1) (tail s))))))))
	   (display "}"))
	  (else (write s)))))

(define *show-stream* 20)

(define (stream? thing)
  (or (eq? thing the-empty-stream)
      (and (pair? thing) (promise? (cdr thing)))))

(define promise?
  (if (procedure? (delay 3))
      procedure?		;In Scheme 48.
      (lambda (x)
	(not (or (pair? x)
		 (null? x)
		 (number? x)
		 (symbol? x)
		 (boolean? x)
		 ;; etc. etc. etc.
		 )))))
